from dataclasses import dataclass
from typing import Dict, List

from tq42.exception_handling import handle_generic_sdk_errors
from tq42.utils.utils import get_hw_configurations

# this import is also important for re-export!
from com.terraquantum.experiment.v1.experimentrun.experiment_run_pb2 import (
    HardwareProto,
)


@dataclass
class HardwareConfig:
    name: str
    cpu: str
    memory: str
    storage: str
    gpu: str


class Compute:
    hardware: HardwareProto
    config: HardwareConfig

    @handle_generic_sdk_errors
    def __init__(self, hardware: HardwareProto = HardwareProto.HARDWARE_UNSPECIFIED):
        self.hardware = hardware
        hardware_name = HardwareProto.Name(hardware)
        configuration = self._find_matching_hardware(hardware_name)
        self.config = HardwareConfig(
            name=hardware_name,
            cpu=configuration.get("CPUs", ""),
            memory=configuration.get("Memory", ""),
            storage=configuration.get("Storage", ""),
            gpu=configuration.get("GPU", ""),
        )

    def __repr__(self) -> str:
        return f"<Compute Id={self.hardware} Name={HardwareProto.Name(self.hardware)}>"

    def __str__(self) -> str:
        return str(self.show_details())

    @staticmethod
    def _find_matching_hardware(hardware_name) -> Dict[str, str]:
        data = get_hw_configurations()
        if hardware_name in data:
            return data.get(hardware_name)

        return {}

    @handle_generic_sdk_errors
    def show_details(self) -> HardwareConfig:
        """
        Show details of compute configurations.

        For details, see
        https://terra-quantum-tq42sdk-docs.readthedocs-hosted.com/en/latest/Python_Developer_Guide/Selecting_Compute_Resources.html#viewing-available-compute-resources-and-configuration-details
        """
        return self.config


@handle_generic_sdk_errors
def list_all() -> List[Compute]:
    """
    Show available compute configurations.

    For details, see
    https://terra-quantum-tq42sdk-docs.readthedocs-hosted.com/en/latest/Python_Developer_Guide/Selecting_Compute_Resources.html#viewing-available-compute-resources-and-configuration-details
    """
    return [Compute(hardware=val) for key, val in HardwareProto.items() if val != 0]



import unittest

from tq42.cli.parsers.tq42parser import parse_args
from tq42.cli.tq42_all import tq42_all
from tq42.functional_tests.functional_test_config import FunctionalCLITestConfig


class TestFunctionalInvokeCliComputeGroup(unittest.TestCase, FunctionalCLITestConfig):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_compute_list(self):
        args = parse_args(["compute", "list"])
        result = tq42_all(self.client, args)
        self.assertIsNotNone(result)

    def test_compute_show_details(self):
        args = parse_args(["compute", "show-details"])
        result = tq42_all(self.client, args)
        self.assertIsNotNone(result)

    def test_compute_show_details_small(self):
        args = parse_args(["compute", "show-details", "--compute=SMALL"])
        result = tq42_all(self.client, args)
        self.assertIsNotNone(result)

    def test_compute_show_details_small_gpu(self):
        args = parse_args(["compute", "show-details", "--compute=SMALL_GPU"])
        result = tq42_all(self.client, args)
        self.assertIsNotNone(result)

    def test_compute_show_details_large(self):
        args = parse_args(["compute", "show-details", "--compute=LARGE"])
        result = tq42_all(self.client, args)
        self.assertIsNotNone(result)

    def test_compute_show_details_large_gpu(self):
        args = parse_args(["compute", "show-details", "--compute=LARGE_GPU"])
        result = tq42_all(self.client, args)
        self.assertIsNotNone(result)



from tq42.cli import (
    organization_group as org_group,
    project_group as proj_group,
    experiment_run_group as exp_run_group,
    experiment_group as exp_group,
    compute_group as compute_group,
    project_dataset_group as proj_data_group,
)

from tq42.utils import environment_utils as env_group
from typing import Optional
from tq42.client import TQ42Client


def tq42_all(client: TQ42Client, args) -> Optional[str]:
    if args.group == "auth" and args.command == "login":
        client.login()
    elif args.group == "exp" and args.command == "run":
        return exp_run_group.experiment_run_group(client, args)
    elif args.group == "exp":
        return exp_group.experiment_group(client, args)
    elif args.group == "proj" and args.command == "dataset":
        return proj_data_group.project_dataset_group(client, args)
    elif args.group == "proj":
        return proj_group.project_group(client, args)
    elif args.group == "org":
        return org_group.organization_group(client, args)
    elif args.group == "compute":
        return compute_group.compute_group(args)
    elif args.group == "env":
        return env_group.environment_group(client=client, args=args)



from __future__ import annotations

import time
from typing import Optional, List

from tq42.client import TQ42Client
from tq42.exception_handling import handle_generic_sdk_errors
from tq42.compute import HardwareProto
from tq42.algorithm import AlgorithmProto
from tq42.exceptions import ExperimentRunCancelError, ExceedRetriesError
from tq42.utils import utils

from com.terraquantum.experiment.v1.experimentrun.experiment_run_pb2 import (
    ExperimentRunStatusProto,
)
from com.terraquantum.experiment.v3alpha1.experimentrun.experiment_run_pb2 import (
    ExperimentRunProto,
)
from com.terraquantum.experiment.v3alpha1.experimentrun.cancel_experiment_run_request_pb2 import (
    CancelExperimentRunRequest,
)
from com.terraquantum.experiment.v3alpha1.experimentrun.get_experiment_run_request_pb2 import (
    GetExperimentRunRequest,
)
from com.terraquantum.experiment.v3alpha1.experimentrun.list_experiment_runs_pb2 import (
    ListExperimentRunsRequest,
)
from com.terraquantum.experiment.v3alpha1.experimentrun.list_experiment_runs_pb2 import (
    ListExperimentRunsResponse,
)


class ExperimentRun:
    id: str
    client: TQ42Client
    data: ExperimentRunProto

    """
    Class to run experiments and view results
    """

    def __init__(
        self, client: TQ42Client, id: str, data: Optional[ExperimentRunProto] = None
    ):
        self.client = client
        self.id = id

        if data:
            self.data = data
        else:
            self.data = self._get_data()

    def __repr__(self) -> str:
        return f"<ExperimentRun Id={self.id}>"

    def __str__(self) -> str:
        return str(self.data)

    @handle_generic_sdk_errors
    def _get_data(self) -> ExperimentRunProto:
        """
        Gets a specific experiment run by id
        """
        get_exp_run_request = GetExperimentRunRequest(experiment_run_id=self.id)

        res = self.client.experiment_run_client.GetExperimentRun(
            request=get_exp_run_request, metadata=self.client.metadata
        )

        return res

    @staticmethod
    def from_proto(client: TQ42Client, msg: ExperimentRunProto) -> ExperimentRun:
        """
        Creates ExperimentRun instance from a protobuf message.
        """
        return ExperimentRun(client=client, id=msg.id, data=msg)

    @staticmethod
    @handle_generic_sdk_errors
    def create(
        client: TQ42Client,
        algorithm: AlgorithmProto,
        experiment_id: str,
        compute: HardwareProto,
        parameters: dict,
    ) -> ExperimentRun:
        """
        Create an experiment run.

        For details, see
        https://terra-quantum-tq42sdk-docs.readthedocs-hosted.com/en/latest/Python_Developer_Guide/Submitting_and_Monitoring_a_Run.html#submitting-an-experiment-run
        """
        create_exp_run_request = utils.dynamic_create_exp_run_request(
            parameters=parameters,
            algo=algorithm,
            exp_id=experiment_id,
            hardware=compute,
        )

        res: ExperimentRunProto = client.experiment_run_client.CreateExperimentRun(
            request=create_exp_run_request, metadata=client.metadata
        )

        client.exp_run_id = utils.get_id(res).strip().replace('"', "")
        return ExperimentRun.from_proto(client=client, msg=res)

    @handle_generic_sdk_errors
    def check(self) -> ExperimentRun:
        """
        Monitor run status.

        For details, see
        https://terra-quantum-tq42sdk-docs.readthedocs-hosted.com/en/latest/Python_Developer_Guide/Submitting_and_Monitoring_a_Run.html#monitoring-an-experiment-run
        """
        self.data = self._get_data()
        return self

    @handle_generic_sdk_errors
    def poll(
        self, tries=1000, initial_delay=1.0, delay=1.0, backoff=1.0
    ) -> ExperimentRun:
        """
        Monitor an experiment run until it completes, then automatically display the results (if there are no errors).

        For details, see
        https://terra-quantum-tq42sdk-docs.readthedocs-hosted.com/en/latest/Python_Developer_Guide/Submitting_and_Monitoring_a_Run.html#monitoring-an-experiment-run
        """
        time.sleep(initial_delay)

        for _ in range(tries):
            self.data = self._get_data()
            if self.data.status in [
                ExperimentRunStatusProto.COMPLETED,
                ExperimentRunStatusProto.CANCELLED,
                ExperimentRunStatusProto.FAILED,
            ]:
                return self

            time.sleep(delay)
            delay *= backoff

        raise ExceedRetriesError(tries=tries)

    @handle_generic_sdk_errors
    def cancel(self) -> ExperimentRun:
        """
        Cancel a run that is QUEUED, PENDING, or RUNNING.

        For details, see
        https://terra-quantum-tq42sdk-docs.readthedocs-hosted.com/en/latest/Python_Developer_Guide/Submitting_and_Monitoring_a_Run.html#cancelling-an-experiment-run
        """
        try:
            cancel_exp_runs_response = CancelExperimentRunRequest(
                experiment_run_id=self.id
            )
            self.client.experiment_run_client.CancelExperimentRun(
                request=cancel_exp_runs_response, metadata=self.client.metadata
            )
            return self
        except Exception:
            raise ExperimentRunCancelError()


@handle_generic_sdk_errors
def list_all(client: TQ42Client, experiment_id: str) -> List[ExperimentRun]:
    """
    List all the runs within an experiment you have permission to view.

    For details, see
    https://terra-quantum-tq42sdk-docs.readthedocs-hosted.com/en/latest/Python_Developer_Guide/Setting_Up_Your_Environment.html#list-all-runs-within-an-experiment
    """
    list_exp_run_request = ListExperimentRunsRequest(experiment_id=experiment_id)

    res: ListExperimentRunsResponse = client.experiment_run_client.ListExperimentRuns(
        request=list_exp_run_request, metadata=client.metadata
    )
    # TODO: It seems like currently the API returns `experiment_runs` instead of `experimentRuns` as in the protobufs
    return [
        ExperimentRun.from_proto(client=client, msg=experiment_run)
        for experiment_run in res.experiment_runs
    ]



from __future__ import annotations

from typing import Optional, List

from google.protobuf import empty_pb2

from tq42.project import Project
from tq42.exception_handling import handle_generic_sdk_errors

from com.terraquantum.organization.v1.organization.organization_pb2 import (
    OrganizationProto,
    ListOrganizationsResponse,
)
from com.terraquantum.organization.v1.organization.get_organization_request_pb2 import (
    GetOrganizationRequest,
)

from typing import TYPE_CHECKING

# only import the stuff for type hints -> avoid circular imports
if TYPE_CHECKING:
    from tq42.client import TQ42Client


class Organization:
    """
    Class to manage organization
    """

    client: TQ42Client
    id: str
    data: OrganizationProto

    def __init__(
        self, client: TQ42Client, id: str, data: Optional[OrganizationProto] = None
    ) -> None:
        self.client = client
        self.id = id

        if data:
            self.data = data
        else:
            self.data = self._get()

    def __repr__(self) -> str:
        return f"<Organization Id={self.id} Name={self.data.name}>"

    def __str__(self) -> str:
        return str(self.data)

    @handle_generic_sdk_errors
    def _get(self) -> OrganizationProto:
        """
        Gets the information about the provided organization
        """
        get_org_request = GetOrganizationRequest(id=self.id)
        res: OrganizationProto = self.client.organization_client.GetOrganization(
            request=get_org_request, metadata=self.client.metadata
        )
        return res

    @staticmethod
    def from_proto(client: TQ42Client, msg: OrganizationProto) -> Organization:
        """
        Creates organization instance from a protobuf message.
        """
        return Organization(client=client, id=msg.id, data=msg)

    @handle_generic_sdk_errors
    def set(self) -> Organization:
        """
        Sets the given organization as the default.
        """
        project = Project.get_default(client=self.client, organization_id=self.id)
        if project:
            project.set()
            return self

        raise KeyError()

    @staticmethod
    @handle_generic_sdk_errors
    def get_default_org(client: TQ42Client) -> Optional[Organization]:
        """
        Gets the default organization for this user based on the default_org field
        """
        org_list = list_all(client=client)
        for org in org_list:
            if org.data.default_org and org.data.default_org is True:
                return org
        return None


@handle_generic_sdk_errors
def list_all(client: TQ42Client) -> List[Organization]:
    """
    List all the organizations you have permission to view.

    For details, see
    https://terra-quantum-tq42sdk-docs.readthedocs-hosted.com/en/latest/Python_Developer_Guide/Setting_Up_Your_Environment.html#list-all-organizations
    """
    empty = empty_pb2.Empty()
    res: ListOrganizationsResponse = (
        client.organization_client.ListAssignedOrganizations(
            request=empty, metadata=client.metadata
        )
    )
    return [
        Organization.from_proto(client=client, msg=msg) for msg in res.organizations
    ]



from tq42.cli.parsers.params_checker import check_params
from tq42.dataset import Dataset, list_all


def project_dataset_group(client, args):
    if args.subcommand == "create":
        return proj_dataset_create(client, args)

    elif args.subcommand == "list":
        return proj_dataset_list(client, args)

    elif args.subcommand == "get":
        return proj_dataset_get(client, args)

    elif args.subcommand == "export":
        return proj_dataset_export(client, args)


def proj_dataset_list(client, args):
    check_params("proj dataset list", args)
    return [dataset.data for dataset in list_all(client=client, project_id=args.proj)]


def proj_dataset_create(client, args):
    check_params("proj dataset create", args)
    return Dataset.create(
        client=client,
        project_id=args.proj,
        name=args.name,
        description=args.desc,
        url=args.url,
        sensitivity=args.sensitivity.upper(),
    ).data


def proj_dataset_get(client, args):
    check_params("proj dataset get", args)
    return Dataset(client=client, id=args.dataset).data


def proj_dataset_export(client, args):
    check_params("proj dataset export", args)
    dataset = Dataset(client=client, id=args.dataset)
    dataset.export(directory_path=args.directory_path)



import tq42.cli.cli_functions as cli
from tq42.cli.parsers.params_checker import check_params


def organization_group(client, args):
    res = None
    if args.command == "set":
        res = org_set(client, args)

    elif args.command == "list":
        res = cli.list_orgs(client)

    return res


def org_set(client, args):
    check_params("org set", args)
    res = cli.set_org(client, args.org)

    return res



from __future__ import annotations
from typing import Optional

from tq42.organization import Organization, list_all as list_all_organizations
from tq42.project import Project, list_all as list_all_projects
from tq42.utils import dirs, file_handling
from tq42.utils.utils_for_cache import clear_cache

from typing import TYPE_CHECKING

# only import the stuff for type hints -> avoid circular imports
if TYPE_CHECKING:
    from tq42.client import TQ42Client


def environment_group(client: TQ42Client, args) -> Optional[str]:
    if args.command == "print":
        return environment_print()

    elif args.command == "default":
        return environment_default_set(client=client)

    elif args.command == "clear":
        return environment_clear()


def environment_print() -> None:
    content = file_handling.read_file(dirs.cache_file())
    print(content)


def environment_clear() -> None:
    clear_cache()
    environment_print()


def environment_default_set(
    client: TQ42Client, organization: Optional[Organization] = None
) -> str:
    if organization is None:
        organization = Organization.get_default_org(client=client)

    organization.set()

    content = file_handling.read_file(dirs.cache_file())
    return content


def get_default_org(client: TQ42Client) -> Optional[Organization]:
    org_list = list_all_organizations(client=client)
    for org in org_list:
        if org.data.default_org and org.data.default_org is True:
            return org
    return None


def get_default_proj(client: TQ42Client, org_id: str) -> Optional[Project]:
    # get default proj id
    project_list = list_all_projects(client=client, organization_id=org_id)

    for proj in project_list:
        if proj.data.default_project and proj.data.default_project is True:
            return proj

    return None



from tq42.exceptions import InvalidInputCliError


def check_params(command, args):
    if command == "exp run create":
        if (
            args.algorithm is None
            or args.compute is None
            or args.parameters is None
            or args.exp is None
        ):
            example_command = "Example usage: tq42 exp run create --exp 98ccb1d2-a3d0-48c8-b172-022f6db9be01  --compute small --algorithm TETRA_OPT --parameters \"{'parameters': {'dimensionality':6,'maximal_rank' :1, 'points_number': 1, 'quantization' : True , 'tolerance':3.9997,  'grid': [1,2,3], 'upper_limits':[1,2,3,4,6,6], 'lower_limits': [0,0,0,0,0,0] , 'objective_function':'https://terraquantum.swiss', 'iteration_number': 1}, 'inputs': {}}\" \n"
            raise SystemExit(InvalidInputCliError(msg=example_command))

    elif command == "org set":
        if args.org is None:
            example_command = (
                "Example usage: tq42 org set 5bac0b60-48d0-45cd-bf0a-39505b058106 \n\n"
            )
            raise SystemExit(InvalidInputCliError(msg=example_command))

    elif command == "exp set-friendly-name":
        if args.name is None or args.exp is None or args.exp == "":
            example_command = 'Example usage: tq42 exp set-friendly-name "FRIENDLY_NAME" --exp="EXP_ID"\n'
            raise SystemExit(InvalidInputCliError(msg=example_command))

    elif command == "exp run list":
        if not args.exp:
            example_command = "Example usage: tq42 exp run list --exp 98ccb1d2-a3d0-48c8-b172-022f6db9be01\n"
            raise SystemExit(InvalidInputCliError(msg=example_command))

    elif command == "proj list":
        if args.proj is None:
            example_command = "Example usage: tq42 proj dataset list --proj 98ccb1d2-a3d0-48c8-b172-022f6db9be01\n"
            raise SystemExit(InvalidInputCliError(msg=example_command))

    elif command == "proj dataset create":
        if (
            args.proj is None
            or args.name is None
            or args.desc is None
            or args.url is None
            or args.sensitivity is None
        ):
            example_command = 'Example usage: tq42 proj dataset create --proj 98ccb1d2-a3d0-48c8-b172-022f6db9be01 --name "Example Dataset Name" --desc "Example Description"  --url "https://mydata.com/drive/my-drive" --sensitivity "confidential" \n'
            raise SystemExit(InvalidInputCliError(msg=example_command))

    elif command == "proj dataset list":
        if args.proj is None:
            example_command = "Example usage: tq42 proj dataset list --proj 98ccb1d2-a3d0-48c8-b172-022f6db9be01\n"
            raise SystemExit(InvalidInputCliError(msg=example_command))

    elif command == "proj dataset get":
        if args.dataset is None:
            example_command = "Example usage: tq42 proj dataset get 98ccb1d2-a3d0-48c8-b172-022f6db9be01\n"
            raise SystemExit(InvalidInputCliError(msg=example_command))

    elif command == "proj dataset export":
        if args.dataset is None or args.directory_path is None:
            example_command = "Example usage: tq42 proj dataset export 98ccb1d2-a3d0-48c8-b172-022f6db9be01 /Users/user1/Downloads\n"
            raise SystemExit(InvalidInputCliError(msg=example_command))



from dataclasses import asdict

import yaml

from tq42.compute import Compute, list_all, HardwareProto


def compute_group(args):
    if args.command == "list":
        return compute_list()

    elif args.command == "show-details":
        return compute_show_details(args)


def compute_list() -> str:
    details = [asdict(hw.show_details()) for hw in list_all()]
    return yaml.dump(details, default_flow_style=False, sort_keys=False)


def compute_show_details(args):
    # noinspection PyBroadException
    try:
        hardware = args.compute.upper()
        compute = Compute(hardware=HardwareProto.Value(hardware))
        details = asdict(compute.show_details())
        return yaml.dump(details, default_flow_style=False, sort_keys=False)
    except Exception:
        options_list = [key for key, val in HardwareProto.items() if val != 0]
        options = ", ".join(options_list)
        example = f"tq42 compute show-details --compute=[{options}]"
        return "Invalid command. \nUsage: " + example



import tq42.utils.dirs as dirs

text_files_dir = dirs.text_files_dir()
unauthenticated_error_file = dirs.full_path(text_files_dir, "unauthenticated_error.txt")
insufficient_permission_errors_file = dirs.full_path(
    text_files_dir, "insufficient_permission_error.txt"
)
no_default_error_file = dirs.full_path(text_files_dir, "no_default_error.txt")
invalid_arguments_error_file = dirs.full_path(
    text_files_dir, "invalid_arguments_error.txt"
)
local_permission_error_file = dirs.full_path(
    text_files_dir, "local_permission_error.txt"
)



from abc import ABCMeta, abstractmethod
from typing import List

from google.protobuf.json_format import MessageToJson
from google.protobuf.message import Message

from com.terraquantum.experiment.v1.experimentrun.experiment_run_pb2 import (
    ExperimentRunProto,
    ExperimentRunStatusProto,
)

from ...project import Project
from tq42.algorithm import AlgorithmProto
from tq42.compute import HardwareProto


class ItemWithIDFormatter(metaclass=ABCMeta):
    @abstractmethod
    def _prefix(self):
        pass

    @abstractmethod
    def _get_items(self, list_items):
        pass

    def format_by_id(self, id):
        return '{}="{}"'.format(self._prefix(), id)

    def format(self, item):
        return self.format_by_id(item.id)

    def format_by_list_object(self, list_items):
        return [self.format(item) for item in self._get_items(list_items)]


class ItemWithNameFormatterResourceClasses(ItemWithIDFormatter, metaclass=ABCMeta):
    def format_by_name_and_id(self, name, id):
        if name is None:
            return self.format_by_id(id)
        else:
            return '{}="{}" ({})'.format(self._prefix(), name, id)

    def format(self, item):
        return self.format_by_name_and_id(item.data.name, item.id)


class OrganizationFormatter(ItemWithIDFormatter):
    def _prefix(self):
        return "org"

    def format_by_project(self, proj: Project):
        return self.format_by_id(proj.data.organization_id)

    def _get_items(self, list_items):
        return list_items


class ProjectFormatter(ItemWithNameFormatterResourceClasses):
    def _prefix(self):
        return "proj"

    def _get_items(self, list_items):
        return list_items


class ExperimentFormatter(ItemWithNameFormatterResourceClasses):
    def _prefix(self):
        return "exp"

    def _get_items(self, list_items):
        return list_items


class ExpRunFormatter(ItemWithIDFormatter):
    def _prefix(self):
        return "run"

    def _get_items(self, list_items):
        return list_items

    def run_created_lines(self, run: ExperimentRunProto):
        status_line = 'status="{}"'.format(ExperimentRunStatusProto.Name(run.status))
        return [self.format(run), status_line]

    def run_checked_lines(self, run: ExperimentRunProto):
        algo_line = 'algorithm="{}"'.format(AlgorithmProto.Name(run.algorithm))
        compute_line = 'compute="{}"'.format(HardwareProto.Name(run.hardware))

        json_msg = (
            MessageToJson(run.result) if isinstance(run.result, Message) else "ERROR"
        )
        result_line = 'result="{}"'.format(json_msg)

        error_message = (
            'error_message="{}"'.format(run.error_message)
            if run.error_message != ""
            else ""
        )
        return self.run_created_lines(run) + [
            algo_line,
            compute_line,
            result_line,
            error_message,
        ]


org_formatter = OrganizationFormatter()
proj_formatter = ProjectFormatter()
exp_formatter = ExperimentFormatter()
run_formatter = ExpRunFormatter()


def set_proj_lines(proj: Project) -> List[str]:
    org_line = org_formatter.format_by_project(proj)
    proj_line = proj_formatter.format(proj)
    return [org_line, proj_line]



import tq42.cli.cli_functions as cli


def project_group(client, args):
    res = None
    if args.command == "list":
        res = cli.list_proj_by_org(client, args.org)
    elif args.command == "show":
        res = cli.proj_show(client)
    elif args.command == "set":
        res = cli.set_project(client, args.proj)
    elif args.command == "set-friendly-name":
        res = proj_set_friendly_name(client, args)

    return res


def proj_set_friendly_name(client, args):
    if args.name is None or args.proj is None:
        example_command = 'Example usage: tq42 proj set-friendly-name "FRIENDLY_NAME" --proj="PROJ_ID"\n'
        print(example_command)
        return

    return cli.proj_update(client, args)



import unittest
from tq42.cli import tq42_help


class MockParser:
    result = None

    def add_argument(self, first_flag, second_flag=None, help=None, metavar=None):
        if second_flag is not None:
            short_flag = first_flag
            flag = second_flag
        else:
            short_flag = None
            flag = first_flag
        self.result = {
            "flag": flag,
            "short_flag": short_flag,
            "help": help,
            "metavar": metavar,
        }


class TestTQ42HelpFlags(unittest.TestCase):
    def setUp(self):
        self.org_flag = {
            "metavar": "ORG_ID",
            "text": "Attribute the command to a particular organization",
            "flag": "--org",
        }
        self.visibility_flag = {
            "text": "Specify the visibility of an attribute",
            "flag": "--visibility",
        }
        self.version_flag = {
            "text": "Display version number",
            "flag": "--version",
            "short_flag": "-V",
        }
        self.help_list = [self.org_flag, self.visibility_flag, self.version_flag]

    def tearDown(self):
        pass

    def test_default_flags(self):
        d = tq42_help.default_help_flags
        self.assertIn("org", d)
        h = d["org"]
        self.assertEqual("--org", h.flag)
        self.assertIsNone(h.short_flag)
        self.assertEqual("Attribute the command to a particular organization", h.text)
        self.assertEqual("ORG_ID", h.metavar)

    def test_from_dict_with_metavar(self):
        h = tq42_help.HelpFlag.from_dict(self.org_flag)
        self.assertEqual("ORG_ID", h.metavar)
        self.assertEqual("Attribute the command to a particular organization", h.text)
        self.assertEqual("--org", h.flag)
        self.assertIsNone(h.short_flag)

    def test_from_dict_without_metavar(self):
        h = tq42_help.HelpFlag.from_dict(self.visibility_flag)
        self.assertIsNone(h.metavar)
        self.assertEqual("Specify the visibility of an attribute", h.text)
        self.assertEqual("--visibility", h.flag)
        self.assertIsNone(h.short_flag)

    def test_from_dict_with_short_flag(self):
        h = tq42_help.HelpFlag.from_dict(self.version_flag)
        self.assertIsNone(h.metavar)
        self.assertEqual("Display version number", h.text)
        self.assertEqual("--version", h.flag)
        self.assertEqual("-V", h.short_flag)

    def test_flag_list_to_dict(self):
        d = tq42_help.HelpFlag.flag_list_to_dict(self.help_list)
        expected = ["org", "version", "visibility"]
        actual = list(d.keys())
        actual.sort()
        self.assertEqual(expected, actual)

    def test_add_flag(self):
        parser = MockParser()

        h = tq42_help.HelpFlag.from_dict(self.org_flag)
        h.add_argument(parser)
        actual = parser.result
        expected = {
            "flag": "--org",
            "short_flag": None,
            "help": "Attribute the command to a particular organization",
            "metavar": "ORG_ID",
        }
        self.assertEqual(expected, actual)

        h = tq42_help.HelpFlag.from_dict(self.visibility_flag)
        h.add_argument(parser)
        actual = parser.result
        expected = {
            "flag": "--visibility",
            "short_flag": None,
            "help": "Specify the visibility of an attribute",
            "metavar": None,
        }
        self.assertEqual(expected, actual)

        h = tq42_help.HelpFlag.from_dict(self.version_flag)
        h.add_argument(parser)
        actual = parser.result
        expected = {
            "flag": "--version",
            "short_flag": "-V",
            "help": "Display version number",
            "metavar": None,
        }
        self.assertEqual(expected, actual)



import tq42.cli.cli_functions as cli
import tq42.utils.utils_for_cache as cache_utils
from tq42.cli.parsers.params_checker import check_params


def experiment_group(client, args):
    if args.command == "list":
        return exp_list_by_proj(client, args)

    elif args.command == "set-friendly-name":
        return exp_set_friendly_name(client, args)


def exp_list_by_proj(client, args):
    # TODO: Need to display error message to stderr when no proj is specified or set.
    #  Should do this by raising an exception and displaying from main function

    error_msg = "No project id is set. \nExample: tq42 exp list --proj b0edfd26-0817-4818-a278-17ef6c14e3a5"

    if args.proj is None:
        proj = cache_utils.get_current_value("proj")
        if proj == "":
            return error_msg
        else:
            return cli.list_exp_by_proj(client, proj)

    return cli.list_exp_by_proj(client, args.proj)


def exp_set_friendly_name(client, args):
    check_params("exp set-friendly-name", args)
    return cli.exp_update(client, args)



import unittest

from tq42.compute import Compute, list_all, HardwareProto


class TestCompute(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_compute_list_all(self):
        all_details = list_all()
        self.assertIsNotNone(all_details)
        self.assertLess(3, len(all_details))
        self.assertNotIn(
            "HARDWARE_UNSPECIFIED", [hw.show_details().name for hw in all_details]
        )

    def test_compute_show_details(self):
        details = Compute(hardware=HardwareProto.SMALL).show_details()
        self.assertEqual(details.name, HardwareProto.Name(HardwareProto.SMALL))

    def test_compute_invalid_type(self):
        with self.assertRaises(AttributeError):
            Compute(hardware=HardwareProto.NOT_A_HARDWARE)



import unittest

from tq42.cli.parsers.tq42parser import parse_args
from tq42.cli.tq42_all import tq42_all
from tq42.functional_tests.functional_test_config import FunctionalCLITestConfig


class TestFunctionalInvokeCliOrganizationGroup(
    unittest.TestCase, FunctionalCLITestConfig
):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_org_list(self):
        args = parse_args(["org", "list"])
        result = tq42_all(self.client, args)
        self.assertIsNotNone(result)

    def test_org_set(self):
        args = parse_args(["org", "set", self.org])
        result = tq42_all(self.client, args)
        self.assertIsNotNone(result)



import unittest

from tq42 import __main__ as main
from tq42.cli import (
    cli_functions,
    compute_group,
    experiment_group,
    experiment_run_group,
    organization_group,
    project_group,
    tq42_all,
    tq42_help,
)
from tq42 import algorithm, organization
from tq42.cli.output_format import formatter
from tq42.cli.parsers import params_checker, tq42parser
from tq42.utils import dirs, utils
from tq42.client import TQ42Client


class TestImports(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_main(self):
        self.assertIsNotNone(main)

    def test_tq42_imports(self):
        self.assertIsNotNone(algorithm)
        self.assertIsNotNone(organization)

    def test_utils_imports(self):
        self.assertIsNotNone(TQ42Client)
        self.assertIsNotNone(dirs)
        self.assertIsNotNone(utils)

    def test_cli_imports(self):
        self.assertIsNotNone(cli_functions)
        self.assertIsNotNone(compute_group)
        self.assertIsNotNone(experiment_group)
        self.assertIsNotNone(experiment_run_group)
        self.assertIsNotNone(organization_group)
        self.assertIsNotNone(project_group)
        self.assertIsNotNone(tq42_all)
        self.assertIsNotNone(tq42_help)

    def test_parser_imports(self):
        self.assertIsNotNone(params_checker)
        self.assertIsNotNone(tq42parser)

    def test_formatter_imports(self):
        self.assertIsNotNone(formatter)



import logging
import traceback

from grpc import StatusCode
from grpc._channel import _InactiveRpcError as InactiveRpcError

from tq42 import exceptions
from functools import wraps

from typing import Any, Callable, TypeVar, cast

F = TypeVar("F", bound=Callable[..., Any])


def handle_generic_sdk_errors(func: F) -> F:
    @wraps(func)
    def wrapped(*args, **kwargs):
        try:
            return func(*args, **kwargs)

        except InactiveRpcError as e:
            status_code = e.code()
            if status_code == StatusCode.PERMISSION_DENIED:
                raise exceptions.PermissionDeniedError() from None

            if status_code == StatusCode.UNAUTHENTICATED:
                raise exceptions.UnauthenticatedError() from None

            if status_code in [
                StatusCode.INVALID_ARGUMENT,
                StatusCode.NOT_FOUND,
                StatusCode.UNKNOWN,
            ]:
                # offending command will be second from the last
                # last line is: traceback.extract_stack()
                index = -2 if len(traceback.extract_stack()) > 1 else 0
                raise exceptions.InvalidArgumentError(
                    command=traceback.extract_stack()[index].line, details=e.details()
                ) from None

            raise e

        except KeyError:
            raise exceptions.NoDefaultError(
                command=traceback.extract_stack()[0].line
            ) from None
        except exceptions.NoMatchingAttributeError as e:
            raise exceptions.InvalidArgumentError(
                command=traceback.extract_stack()[0].line, details=e.details
            )
        except FileExistsError as e:
            raise exceptions.InvalidArgumentError(
                command=traceback.extract_stack()[0].line,
                details=f"File {e} already exists",
            )
        except (
            exceptions.AuthenticationError,
            exceptions.PermissionDeniedError,
            exceptions.UnauthenticatedError,
            exceptions.InvalidArgumentError,
            exceptions.NoDefaultError,
            exceptions.ExceedRetriesError,
            exceptions.LocalPermissionError,
            exceptions.ExperimentRunCancelError,
        ) as e:
            raise e from None
        except (PermissionError, OSError):
            raise exceptions.LocalPermissionError() from None
        except Exception as e:
            # are there generic errors we are not catching here?
            logging.debug(
                "Error {} is raised in SDK and not specifically handled".format(e)
            )
            raise e

    return cast(F, wrapped)



import json
import os
import webbrowser
from datetime import datetime
import grpc
from grpc import aio
import requests
from tq42.utils import dirs, file_handling, utils
from tq42.utils.token_manager import TokenManager
from tq42.exception_handling import handle_generic_sdk_errors
import time

from com.terraquantum.experiment.v3alpha1.experiment import (
    experiment_service_pb2_grpc as pb2_exp_grpc,
)
from com.terraquantum.experiment.v3alpha1.experimentrun import (
    experiment_run_service_pb2_grpc as pb2_exp_run_grpc,
)
from com.terraquantum.organization.v1.organization import (
    organization_service_pb2_grpc as pb2_org_grpc,
)
from com.terraquantum.project.v1.project import (
    project_service_pb2_grpc as pb2_proj_grpc,
)
from com.terraquantum.storage.v1alpha1 import (
    storage_service_pb2_grpc as pb2_data_grpc,
)
from com.terraquantum.channel.v1alpha1 import (
    channel_service_pb2_grpc as pb2_channel_grpc,
)
from tq42.utils.environment_utils import environment_default_set
from tq42.exceptions import AuthenticationError


class ConfigEnvironment:
    """
    URLs determining environment
    """

    def __init__(self, base_url, client_id, scope):
        self.base_url = base_url
        self.client_id = client_id
        self.scope = scope

    @property
    def api_host(self):
        return "api.{}".format(self.base_url)

    @property
    def channels_host(self):
        return "channels.{}".format(self.base_url)

    @property
    def auth_url_token(self):
        return "https://auth.{}/oauth/token".format(self.base_url)

    @property
    def auth_url_code(self):
        return "https://auth.{}/oauth/device/code".format(self.base_url)

    @property
    def audience(self):
        return "https://graphql-gateway.{}/graphql".format(self.base_url)

    @property
    def client_credential_flow_audience(self):
        return "https://api.{}".format(self.base_url)

    @property
    def headers(self):
        return {"Content-Type": "application/x-www-form-urlencoded"}

    @property
    def code_data(self):
        return {
            "client_id": self.client_id,
            "scope": self.scope,
            "audience": self.audience,
        }

    def token_data(self, device_code):
        return {
            "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
            "device_code": device_code,
            "client_id": self.client_id,
        }

    def refresh_token_data(self, refresh_token):
        return {
            "grant_type": "refresh_token",
            "refresh_token": refresh_token,
            "client_id": self.client_id,
        }

    def client_credentials_data(self, client_id, client_secret, audience):
        return {
            "client_id": client_id,
            "client_secret": client_secret,
            "grant_type": "client_credentials",
            "audience": audience,
        }


class TQ42Client(object):
    """
    Visit tq42.com/help for more details on all commands.
    """

    def __call__(self, **kwargs):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def __init__(self, alt_config_file=None):
        # check whether "alt_config_file" is the default filepath, if it is, token.json resides
        # in ~/.tqsdk/ and config.json inside utils/text_files/, if not using default,
        # config.json and token.json will reside in the same path
        # this happens when the user specifies a path during initialization of tq42 (usually for doing tests)

        self.default_config_file = dirs.text_files_dir("config.json")
        self.config_file = self.default_config_file
        self.config_folder = dirs.get_config_folder_path()

        if alt_config_file is not None and not self.is_config_filepath_default(
            alt_config_file
        ):
            self.config_file = alt_config_file
            self.config_folder = os.path.dirname(alt_config_file)

        with open(self.config_file, encoding="utf-8") as f:
            config_data = json.load(f)

        environment = ConfigEnvironment(
            config_data["base_url"], config_data["client_id"], config_data["scope"]
        )

        self.token_manager = TokenManager(environment, self.config_folder)

        self.environment = environment
        self.api_host = environment.api_host
        self.channels_host = environment.channels_host
        self.server_port = 443

        self.exp_run_id = None
        # instantiate a channel
        self.api_channel = grpc.secure_channel(
            self.api_host,
            grpc.ssl_channel_credentials(),
            options=[
                ("grpc.max_receive_message_length", 10_000_000),
            ],
        )
        self.channels_channel = aio.secure_channel(
            self.channels_host, grpc.ssl_channel_credentials()
        )

        # bind the client and the server
        self.organization_client = pb2_org_grpc.OrganizationServiceStub(
            self.api_channel
        )
        self.project_client = pb2_proj_grpc.ProjectServiceStub(self.api_channel)
        self.experiment_client = pb2_exp_grpc.ExperimentServiceStub(self.api_channel)
        self.storage_client = pb2_data_grpc.StorageServiceStub(self.api_channel)
        self.experiment_run_client = pb2_exp_run_grpc.ExperimentRunServiceStub(
            self.api_channel
        )
        self.channel_client = pb2_channel_grpc.ChannelServiceStub(self.channels_channel)
        self.credential_flow_client_id = os.getenv("TQ42_AUTH_CLIENT_ID")
        self.credential_flow_client_secret = os.getenv("TQ42_AUTH_CLIENT_SECRET")

    @handle_generic_sdk_errors
    def login_without_user_interaction(self):
        response = requests.post(
            self.environment.auth_url_token,
            data=self.environment.client_credentials_data(
                client_id=self.credential_flow_client_id,
                client_secret=self.credential_flow_client_secret,
                audience=self.environment.client_credential_flow_audience,
            ),
            headers=self.environment.headers,
        )

        response_json = response.json()
        access_token = response_json.get("access_token")

        if not access_token:
            raise AuthenticationError(
                "No access token can be retrieved from the response."
            )

        self.save_access_token(access_token)

    @handle_generic_sdk_errors
    def login(self):
        if self.credential_flow_client_id and self.credential_flow_client_secret:
            self.login_without_user_interaction()
        else:
            self.login_with_user_interaction()

    def login_with_user_interaction(self):
        """
        This command will open a window in your browser where you must confirm the MFA code, then enter your TQ42
        username and password to authenticate. The authentication validity will keep extending as long as you are
        using it within a 30 day period. To access TQ42 services with Python commands, you need a TQ42 account.
        When running TQ42 Python commands, your environment needs to have access to your TQ42 account credentials.

        For details, see
         https://terra-quantum-tq42sdk-docs.readthedocs-hosted.com/en/latest/README.html#authentication
        """
        # Send the POST request and print the response
        response = requests.post(
            self.environment.auth_url_code,
            data=self.environment.code_data,
            headers=self.environment.headers,
        )

        json_response = response.json()
        print(response.text)
        user_code = json_response["user_code"]
        device_code = json_response["device_code"]
        verification_uri_complete = json_response["verification_uri_complete"]
        interval = json_response["interval"]

        # Print the message to the user
        print(
            f"If a browser does not open, please access this URL: {verification_uri_complete} to login. Also check the code: {user_code}"
        )

        webbrowser.open(verification_uri_complete)

        data_token = self.environment.token_data(device_code)

        while True:
            # Send the POST request to get access token and extract the JSON response
            response_token = requests.post(
                self.environment.auth_url_token,
                data=data_token,
                headers=self.environment.headers,
            )
            response_json = response_token.json()

            refresh_token = response_json.get("refresh_token")
            access_token = response_json.get("access_token")

            # If we received an access token, print it and break out of the loop
            if refresh_token and access_token:
                self.save_access_token(access_token)
                self.save_refresh_token(refresh_token)
                break

            # Otherwise, wait for the specified interval before polling again
            time.sleep(interval)

    def save_access_token(self, access_token: str):
        save_location = utils.save_token(
            service_name="access_token",
            backup_save_path=self.token_file_path,
            token=access_token,
        )

        print(
            f"Authentication is successful, access token is saved in: {save_location}."
        )

        env_set = environment_default_set(client=self)
        print(env_set)

    def save_refresh_token(self, refresh_token: str):
        utils.save_token(
            service_name="refresh_token",
            backup_save_path=self.refresh_token_file_path,
            token=refresh_token,
        )
        current_datetime = datetime.now()
        file_handling.write_to_file(self.timestamp_file_path, current_datetime)

    def is_config_filepath_default(self, config_file):
        return config_file == self.default_config_file

    @property
    def token_file_path(self):
        return self.token_manager.token_file_path

    @property
    def timestamp_file_path(self):
        return self.token_manager.timestamp_file_path

    @property
    def refresh_token_file_path(self):
        return self.token_manager.refresh_token_file_path

    @property
    def metadata(self):
        self.token_manager.renew_expring_token()
        token = utils.get_token(
            service_name="access_token", backup_save_path=self.token_file_path
        )
        return (("authorization", "Bearer " + token),)



from tq42.cli.output_format.formatter import (
    org_formatter,
    proj_formatter,
    exp_formatter,
    run_formatter,
    set_proj_lines,
)
from tq42.organization import list_all as list_all_organizations, Organization
from tq42.project import list_all as list_all_projects, Project
from tq42.experiment import list_all as list_all_experiments, Experiment
from tq42.experiment_run import list_all as list_all_experiment_runs, ExperimentRun

from tq42.client import TQ42Client


def list_orgs(client: TQ42Client) -> str:
    org_list = org_formatter.format_by_list_object(
        list_all_organizations(client=client)
    )
    result = "\n".join(org_list)
    return result


def proj_update(client: TQ42Client, args) -> str:
    proj = args.proj
    friendly_name = args.name
    updated_proj = Project(client=client, id=proj).update(name=friendly_name)
    result = proj_formatter.format(updated_proj)
    return result


def exp_update(client: TQ42Client, args) -> str:
    exp = args.exp
    result = exp_formatter.format(
        Experiment(client=client, id=exp).update(name=args.name)
    )
    return result


def list_proj_by_org(client: TQ42Client, org: str) -> str:
    project_list = proj_formatter.format_by_list_object(
        list_all_projects(client=client, organization_id=org)
    )
    result = "\n".join(project_list)
    return result


def get_exprun(client: TQ42Client, run_id: str) -> str:
    result = run_formatter.run_checked_lines(
        ExperimentRun(client=client, id=run_id).data
    )
    result = "\n".join(result)
    return result


def list_exp_by_proj(client: TQ42Client, proj: str) -> str:
    result = exp_formatter.format_by_list_object(
        list_all_experiments(client=client, project_id=proj)
    )
    result = "\n".join(result)
    return result


def list_expruns(client: TQ42Client, exp: str) -> str:
    result = run_formatter.format_by_list_object(
        list_all_experiment_runs(client=client, experiment_id=exp)
    )
    result = "\n".join(result)
    return result


def cancel_exprun(client: TQ42Client, run_id: str) -> str:
    # if not cancelled it will raise an Exception
    ExperimentRun(client=client, id=run_id).cancel()
    return "The experiment run is cancelled."


def set_org(client: TQ42Client, org_id: str) -> str:
    Organization(client=client, id=org_id).set()
    project = Project.show(client=client)
    result = set_proj_lines(proj=project)
    return "\n".join(result)


def set_project(client: TQ42Client, proj_id: str) -> str:
    project = Project(client=client, id=proj_id).set()
    result = set_proj_lines(project)
    result = "\n".join(result)
    return result


def proj_show(client: TQ42Client) -> str:
    retrieved = Project.show(client=client)
    result = set_proj_lines(retrieved)
    result = "\n".join(result)
    return result


def poll_exprun(client: TQ42Client, run_id: str) -> str:
    try:
        result = ExperimentRun(client=client, id=run_id).poll()
        result = run_formatter.run_checked_lines(result.data)
        result = "\n".join(result)
        return result

    except Exception as e:
        print(e)



import unittest
from unittest.mock import MagicMock, patch
from google.protobuf.json_format import ParseDict
from google.protobuf.timestamp_pb2 import Timestamp
from grpc import StatusCode
from grpc._channel import _InactiveRpcError as InactiveRpcError, _RPCState as RPCState

from com.terraquantum.project.v1.project import (
    project_pb2 as proj_def,
    list_projects_response_pb2 as list_proj_res,
)
from com.terraquantum.organization.v1.organization import (
    organization_pb2 as org_def,
)

from tq42.client import TQ42Client
from tq42.organization import Organization
from tq42.project import Project
from tq42.exceptions import NoDefaultError, InvalidArgumentError


class TestAPI(unittest.TestCase):
    def setUp(self):
        self.client = TQ42Client()

    def tearDown(self):
        pass

    def test_client_structure(self):
        self.assertTrue(hasattr(TQ42Client, "login"))
        self.assertTrue(hasattr(TQ42Client, "__enter__"))
        self.assertTrue(hasattr(TQ42Client, "__exit__"))

    def test_context_manager_behavior(self):
        with TQ42Client() as client:
            # Assert that the __enter__ method is called
            self.assertEqual(client.__enter__(), client)

            # Perform some operations inside the context
            client.metadata

            self.assertNotEqual(client.__exit__(None, None, None), client)

        # Assert that the __exit__ method is called after exiting the context
        self.assertEqual(client.__exit__(None, None, None), None)

    @patch("tq42.project.get_current_value")
    def test_proj_show_no_proj_set(self, get_current_value_mock):
        get_current_value_mock.side_effect = KeyError("key error")
        self.assertRaises(NoDefaultError, Project.show, self.client)

    @patch("tq42.project.get_current_value")
    def test_proj_show_default_proj_set(self, get_current_value_mock):
        project = {
            "id": "this-is-random-but-valid-uuid",
            "organization_id": "this-is-a-random-org-id-1",
            "name": "right project",
            "description": "this is the right project",
            "image_url": "https://random-image.url",
            "state": 1,  # active project state
            "created_at": Timestamp().GetCurrentTime(),
            "created_by": "steven",
            "default_project": True,
        }
        project = ParseDict(project, proj_def.ProjectProto())
        self.assertIsNotNone(project)
        self.client.project_client.GetProject = MagicMock(return_value=project)
        get_current_value_mock.return_value = "this-is-a-random-but-valid-uuid"
        res = Project.show(client=self.client)
        self.assertIsNotNone(res)
        self.assertEqual(1, self.client.project_client.GetProject.call_count)
        self.assertEqual(
            "this-is-a-random-but-valid-uuid",
            self.client.project_client.GetProject.call_args[1]["request"].id,
        )
        self.assertEqual("right project", res.data.name)

    @patch("tq42.project.write_key_value_to_cache")
    @patch("tq42.project.clear_cache")
    def test_proj_set(self, clear_cache_mock, write_key_value_to_cache_mock):
        project = {
            "id": "new-project-uuid-we-wanna-set",
            "organization_id": "this-is-a-random-org-id-1",
            "name": "right project",
            "description": "this is the right project",
            "image_url": "https://random-image.url",
            "state": 1,  # active project state
            "created_at": Timestamp().GetCurrentTime(),
            "created_by": "steven",
            "default_project": True,
        }

        cache = {}

        def add_stuff_to_dict(key: str, value: str) -> None:
            cache[key] = value

        write_key_value_to_cache_mock.side_effect = add_stuff_to_dict

        project = ParseDict(project, proj_def.ProjectProto())
        self.client.project_client.GetProject = MagicMock(return_value=project)
        Project(client=self.client, id="new-project-uuid-we-wanna-set").set()

        self.assertEqual(1, clear_cache_mock.call_count)
        self.assertEqual(2, write_key_value_to_cache_mock.call_count)
        self.assertDictEqual(
            cache,
            {
                "proj": "new-project-uuid-we-wanna-set",
                "org": "this-is-a-random-org-id-1",
            },
        )

    def test_project_not_found(self):
        get_project_mock = MagicMock()
        get_project_mock.side_effect = InactiveRpcError(
            RPCState(
                code=StatusCode.INVALID_ARGUMENT,
                details="no details",
                due=[],
                initial_metadata=[],
                trailing_metadata=[],
            )
        )

        self.client.project_client.GetProject = get_project_mock
        self.assertRaises(
            InvalidArgumentError,
            Project,
            client=self.client,
            id="this-project-id-is-impossible-to-find",
        )

    @patch("tq42.project.write_key_value_to_cache")
    @patch("tq42.project.clear_cache")
    def test_org_set(self, clear_cache_mock, write_key_value_to_cache_mock):
        org = {
            "id": "new-org-uuid-we-wanna-set",
            "owner": "steven",
            "name": "just-so-random-org",
            "description": "random description for org",
            "image_url": "https://random-image.url",
            "state": 1,  # active org state
            "default_org": False,
        }

        projects_dict = {
            "projects": [
                {
                    "id": "75ec987b-bf43-46e5-a0f0-85bc08c9cf18",
                    "organization_id": "new-org-uuid-we-wanna-set",
                    "name": "Ymixer shape optimization",
                    "state": "ACTIVE",
                    "created_by": "e31d88a8-8720-4e4b-b95b-92538899a6ba",
                },
                {
                    "id": "new-proj-uuid-we-wanna-set",
                    "organization_id": "new-org-uuid-we-wanna-set",
                    "name": "b225f5e5-eaa0-47d0-8ef0-7e954da6d681",
                    "state": "ACTIVE",
                    "created_by": "0e02a4b1-28c1-4a33-8995-d39db7de8bb9",
                    "default_project": True,
                },
            ]
        }

        projects = ParseDict(projects_dict, list_proj_res.ListProjectsResponse())
        self.assertIsNotNone(projects)

        cache = {}

        def add_stuff_to_dict(key: str, value: str) -> None:
            cache[key] = value

        write_key_value_to_cache_mock.side_effect = add_stuff_to_dict

        org = ParseDict(org, org_def.OrganizationProto())
        self.client.organization_client.GetOrganization = MagicMock(return_value=org)

        self.client.project_client.ListProjects = MagicMock(return_value=projects)

        Organization(client=self.client, id="new-org-uuid-we-wanna-set").set()

        self.assertEqual(1, clear_cache_mock.call_count)
        self.assertEqual(2, write_key_value_to_cache_mock.call_count)
        expected = {
            "org": "new-org-uuid-we-wanna-set",
            "proj": "new-proj-uuid-we-wanna-set",
        }
        self.assertDictEqual(cache, expected)

    def test_org_not_found(self):
        get_org_mock = MagicMock()
        get_org_mock.side_effect = InactiveRpcError(
            RPCState(
                code=StatusCode.INVALID_ARGUMENT,
                details="no details",
                due=[],
                initial_metadata=[],
                trailing_metadata=[],
            )
        )

        self.client.organization_client.GetOrganization = get_org_mock

        self.assertRaises(
            InvalidArgumentError,
            Organization,
            client=self.client,
            id="this-org-id-is-impossible-to-find",
        )



import argparse
import textwrap

import tq42
from tq42.cli import tq42_help


def parse_args(args):
    if len(args) < 1:
        args = ["--help"]
    parser = argparse.ArgumentParser(
        prog="tq42",
        description=textwrap.dedent(tq42_help.get_commands()),
        formatter_class=argparse.RawTextHelpFormatter,
    )

    config_flag = tq42_help.default_help_flags["config"]
    config_flag.add_argument(parser)

    version_flag = tq42_help.default_help_flags["version"]
    parser.add_argument(
        version_flag.short_flag,
        version_flag.flag,
        help=version_flag.text,
        action="version",
        version=tq42.__version__,
    )

    group_parser = parser.add_subparsers(dest="group")

    add_auth_parser(group_parser)
    add_compute_parser(group_parser)
    add_proj_parser(group_parser)
    add_exp_parser(group_parser)

    add_org_parser(group_parser)
    add_env_parser(group_parser)

    args = parser.parse_args(args)
    return args


def add_auth_parser(group_parser):
    auth_help = tq42_help.Help.lookup_help(["auth"])
    auth_parser = auth_help.add_parser(group_parser)

    auth_subparser = auth_parser.add_subparsers(dest="command")

    login_help = tq42_help.Help.lookup_help("auth login".split())
    login_help.add_parser(auth_subparser)


def add_proj_parser(group_parser):
    proj_help = tq42_help.Help.lookup_help(["proj"])
    proj_parser = proj_help.add_parser(group_parser)

    proj_subparser = proj_parser.add_subparsers(dest="command")

    list_help = tq42_help.Help.lookup_help("proj list".split())
    list_help.add_parser(proj_subparser)
    list_help.add_flag("org")

    show_help = tq42_help.Help.lookup_help("proj show".split())
    show_help.add_parser(proj_subparser)

    set_help = tq42_help.Help.lookup_help("proj set".split())
    proj_set_parser = set_help.add_parser(proj_subparser)
    proj_set_parser.add_argument("proj")

    friendlyname_help = tq42_help.Help.lookup_help("proj set-friendly-name".split())
    proj_friendlyname_parser = friendlyname_help.add_parser(proj_subparser)
    proj_friendlyname_parser.add_argument("name", nargs="?")
    friendlyname_help.add_flag("proj")

    add_proj_dataset_parser(proj_subparser)


def add_proj_dataset_parser(proj_subparser):
    run_help = tq42_help.Help.lookup_help("proj dataset".split())
    proj_subparser = run_help.add_parser(proj_subparser)
    cmd_parser = proj_subparser.add_subparsers(dest="subcommand")

    create_help = tq42_help.Help.lookup_help("proj dataset create".split())
    create_help.add_parser(cmd_parser)
    create_help.add_flag("proj")
    create_help.add_flag("name")
    create_help.add_flag("desc")
    create_help.add_flag("url")
    create_help.add_flag("sensitivity")

    list_help = tq42_help.Help.lookup_help("proj dataset list".split())
    list_help.add_parser(cmd_parser)
    list_help.add_flag("proj")

    get_help = tq42_help.Help.lookup_help("proj dataset get".split())
    dataset_get_parser = get_help.add_parser(cmd_parser)
    dataset_get_parser.add_argument("dataset")

    export_help = tq42_help.Help.lookup_help("proj dataset export".split())
    dataset_export_parser = export_help.add_parser(cmd_parser)
    dataset_export_parser.add_argument("dataset")
    dataset_export_parser.add_argument("directory_path")


def add_org_parser(group_parser):
    org_help = tq42_help.Help.lookup_help(["org"])
    org_parser = org_help.add_parser(group_parser)

    auth_subparser = org_parser.add_subparsers(dest="command")

    set_help = tq42_help.Help.lookup_help("org set".split())
    org_set_parser = set_help.add_parser(auth_subparser)
    org_set_parser.add_argument("org", nargs="?")

    list_help = tq42_help.Help.lookup_help("org list".split())
    list_help.add_parser(auth_subparser)


def add_exp_parser(group_parser):
    exp_help = tq42_help.Help.lookup_help(["exp"])
    exp_parser = exp_help.add_parser(group_parser)

    exp_subparser = exp_parser.add_subparsers(dest="command")

    # tq42 exp list --proj
    list_help = tq42_help.Help.lookup_help("exp list".split())
    list_help.add_parser(exp_subparser)
    list_help.add_flag("proj")

    friendlyname_help = tq42_help.Help.lookup_help("exp set-friendly-name".split())
    exp_setfriendlyname_parser = friendlyname_help.add_parser(exp_subparser)
    exp_setfriendlyname_parser.add_argument("name")
    friendlyname_help.add_flag("exp")

    # tq42 exp run commands
    add_exp_run_parser(exp_subparser)


def add_exp_run_parser(exp_parser):
    run_help = tq42_help.Help.lookup_help("exp run".split())
    exp_parser = run_help.add_parser(exp_parser)

    cmd_parser = exp_parser.add_subparsers(dest="subcommand")

    create_help = tq42_help.Help.lookup_help("exp run create".split())
    create_help.add_parser(cmd_parser)
    create_help.add_flag("exp")
    create_help.add_flag("compute")
    create_help.add_flag("algorithm")
    create_help.add_flag("parameters")

    check_help = tq42_help.Help.lookup_help("exp run check".split())
    check_parser = check_help.add_parser(cmd_parser)
    check_parser.add_argument("run")

    cancel_help = tq42_help.Help.lookup_help("exp run cancel".split())
    cancel_parser = cancel_help.add_parser(cmd_parser)
    cancel_parser.add_argument("run")

    list_help = tq42_help.Help.lookup_help("exp run list".split())
    list_help.add_parser(cmd_parser)
    list_help.add_flag("exp")

    poll_help = tq42_help.Help.lookup_help("exp run poll".split())
    poll_parser = poll_help.add_parser(cmd_parser)
    poll_parser.add_argument("run")


def add_compute_parser(group_parser):
    compute_help = tq42_help.Help.lookup_help(["compute"])
    compute_parser = compute_help.add_parser(group_parser)

    compute_subparser = compute_parser.add_subparsers(dest="command")

    list_help = tq42_help.Help.lookup_help("compute list".split())
    list_help.add_parser(compute_subparser)

    details_help = tq42_help.Help.lookup_help("compute show-details".split())
    details_help.add_parser(compute_subparser)
    details_help.add_flag("compute")


def add_env_parser(group_parser):
    env_parser = group_parser.add_parser("env")

    env_subparser = env_parser.add_subparsers(dest="command")
    env_subparser.add_parser("default")

    env_subparser.add_parser("print")
    env_subparser.add_parser("clear")



import unittest
from tq42.cli import tq42_help


class MockSubparser:
    def add_parser(
        self,
        name,
        description,
        help,
        formatter_class,
    ):
        return {
            "name": name,
            "description": description,
            "help": help,
            "formatter_class": formatter_class,
        }


class TestTQ42Help(unittest.TestCase):
    def setUp(self):
        self.help_tree = {
            "auth": {
                "help": {"text": "Class to manage authentication", "level": "group"},
                "login": {
                    "help": {
                        "text": "Authenticate with this.",
                        "level": "command",
                        "cli_url": "README.html#authentication",
                        "python_url": "README.html#ptyhon_authentication",
                    }
                },
            },
            "help": {"cli_url": "index.html", "level": "tq42"},
        }

    def tearDown(self):
        pass

    def with_urls_test(self, with_urls):
        self.assertEqual("tq42 org list", with_urls.name)
        self.assertEqual("command", with_urls.level)
        self.assertEqual("test.net/cli/help", with_urls.cli_url)
        self.assertEqual("test.net/python/help", with_urls.python_url)
        self.assertEqual("you need help", with_urls.text)

    def without_urls_test(self, no_urls):
        self.assertEqual("tq42 auth", no_urls.name)
        self.assertEqual("group", no_urls.level)
        self.assertIsNone(no_urls.cli_url)
        self.assertIsNone(no_urls.python_url)
        self.assertEqual("authenticate into tq42", no_urls.text)

    @staticmethod
    def get_help_with_urls():
        return tq42_help.Help(
            name="tq42 org list",
            level="command",
            cli_url="test.net/cli/help",
            python_url="test.net/python/help",
            text="you need help",
        )

    @staticmethod
    def get_help_without_urls():
        return tq42_help.Help(
            name="tq42 auth", level="group", text="authenticate into tq42"
        )

    def test_tq42_help_init(self):
        with_urls = self.get_help_with_urls()
        self.with_urls_test(with_urls)

        no_urls = self.get_help_without_urls()
        self.without_urls_test(no_urls)

    def test_tq42_from_dict(self):
        with_urls = tq42_help.Help.from_dict(
            {
                "name": "tq42 org list",
                "level": "command",
                "cli_url": "test.net/cli/help",
                "python_url": "test.net/python/help",
                "text": "you need help",
            }
        )
        self.with_urls_test(with_urls)

        no_urls = tq42_help.Help.from_dict(
            {"name": "tq42 auth", "level": "group", "text": "authenticate into tq42"}
        )
        self.without_urls_test(no_urls)

    def test_display_text(self):
        with_urls = self.get_help_with_urls()
        expected = "you need help\n\nFor details, see test.net/cli/help"
        actual = with_urls.display_help
        self.assertEqual(expected, actual)

        no_urls = self.get_help_without_urls()
        expected = no_urls.text
        actual = no_urls.display_help
        self.assertEqual(expected, actual)

    def test_display_docstring(self):
        with_urls = self.get_help_with_urls()
        expected = "you need help\n\nFor details, see test.net/python/help"
        actual = with_urls.display_docstring
        self.assertEqual(expected, actual)

        no_urls = self.get_help_without_urls()
        expected = no_urls.text
        actual = no_urls.display_docstring
        self.assertEqual(expected, actual)

    def test_walk_tree_tq42(self):
        expected = {"cli_url": "index.html", "level": "tq42", "name": "tq42"}
        actual = tq42_help.Help.walk_tree([], self.help_tree)
        self.assertEqual(expected, actual)

    def test_walk_tree_tq42_auth(self):
        expected = {
            "text": "Class to manage authentication",
            "level": "group",
            "name": "auth",
        }
        actual = tq42_help.Help.walk_tree(["auth"], self.help_tree)
        self.assertEqual(expected, actual)

    def test_walk_tree_tq42_auth_login(self):
        expected = {
            "text": "Authenticate with this.",
            "level": "command",
            "cli_url": "README.html#authentication",
            "python_url": "README.html#ptyhon_authentication",
            "name": "login",
        }
        actual = tq42_help.Help.walk_tree("auth login".split(), self.help_tree)
        self.assertEqual(expected, actual)

    def test_walk_root(self):
        tree = {
            "help": {
                "cli_url": "https://terra-quantum-tq42sdk-docs.readthedocs-hosted.com/en/latest/index.html",
                "level": "tq42",
            }
        }
        h = tq42_help.Help.lookup_help([], tree=tree)
        self.assertEqual(h.name, "tq42")

    def test_default_help_tree(self):
        self.assertEqual("tq42", tq42_help.default_help_tree["help"]["level"])
        global_help_object = tq42_help.Help.lookup_help([])
        self.assertEqual("tq42", global_help_object.level)
        self.assertIsNone(global_help_object.text)
        self.assertIsNotNone(global_help_object.cli_url)
        self.assertIsNone(global_help_object.python_url)

    def test_get_global_help(self):
        global_help_object = tq42_help.Help.lookup_help([])
        cli_url = global_help_object.cli_url
        global_help_text = tq42_help.get_commands()
        self.assertEqual(str, type(global_help_text))
        self.assertIn(cli_url, global_help_text)

    def test_add_parser(self):
        subparser = MockSubparser()
        expected = {
            "name": "list",
            "description": "you need help\n\nFor details, see test.net/cli/help",
            "help": "you need help",
        }
        help = self.get_help_with_urls()
        help.name = "list"
        self.assertIsNone(help.argparser)
        actual = help.add_parser(subparser)
        del actual["formatter_class"]
        self.assertEqual(expected, actual)
        self.assertEqual(help.argparser, actual)
